import argparse, sys

import scripts
from xklb import __version__, utils
from xklb.consts import SC
from xklb.dl_extract import dl_block, dl_download
from xklb.fs_extract import fs_add, fs_update
from xklb.hn_extract import hacker_news_add
from xklb.play_actions import filesystem, listen, read, view, watch
from xklb.playback import playback_next, playback_now, playback_pause, playback_stop
from xklb.praw_extract import reddit_add, reddit_update
from xklb.stats import dlstatus, playlists
from xklb.tabs_actions import tabs
from xklb.tabs_extract import tabs_add
from xklb.tube_extract import tube_add, tube_update
from xklb.utils import log


def usage() -> str:
    return f"""xk media library subcommands (v{__version__})

    local media:
      lb fsadd                 Create a local media database; Add folders
      lb fsupdate              Refresh database: add new files, mark deleted

      lb listen                Listen to local and online media
      lb watch                 Watch local and online media
      lb read                  Read books
      lb view                  View images

      lb bigdirs               Discover folders which take much room
      lb dedupe                Deduplicate local db files
      lb relmv                 Move files/folders while preserving relative paths
      lb mv-list               Reach a target free space by moving data across mount points
      lb scatter               Scatter files across multiple mountpoints (mergerfs balance)

      lb christen              Cleanse files by giving them a new name

    online media:
      lb tubeadd               Create a tube database; Add playlists
      lb tubeupdate            Fetch new videos from saved playlists

      lb redditadd             Create a reddit database; Add subreddits
      lb redditupdate          Fetch new posts from saved subreddits

    downloads:
      lb download              Download media
      lb redownload            Redownload missing media
      lb block                 Prevent downloading specific URLs
      lb merge-dbs             Merge multiple SQLITE files
      lb merge-online-local    Merge local and online metadata

    playback:
      lb now                   Print what is currently playing
      lb next                  Play next file
      lb stop                  Stop all playback
      lb pause                 Pause all playback

    statistics:
      lb playlists             List added playlists
      lb dlstatus              Show download status
      lb usage                 Print mount usage

    browser tabs:
      lb tabsadd               Create a tabs database; Add URLs
      lb tabs                  Open your tabs for the day
      lb surf                  Load browser tabs in a streaming way (stdin)

    mining:
      lb reddit-selftext       db selftext external links -> db media table
      lb pushshift             Convert Pushshift jsonl.zstd -> reddit.db format (stdin)
      lb nfb-films             NFB Director links -> film links (stdin)
      lb nouns                 Unstructured text -> compound nouns (stdin)
      lb hnadd                 Create a hackernews database (this takes a few days)
    """


def print_help(parser) -> None:
    print(usage())
    print(parser.epilog)


subcommands = ["fs", "du"]


def consecutive_prefixes(s):
    prefixes = list(s[:j] for j in range(5, len(s)) if s[:j] and s[:j] not in subcommands)
    subcommands.extend(prefixes)
    return prefixes


def add_parser(subparsers, name, a=None):
    if a is None:
        a = []
    subcommands.extend([name] + a)
    aliases = a + consecutive_prefixes(name) + utils.conform([consecutive_prefixes(a) for a in a])
    return subparsers.add_parser(name, aliases=aliases, add_help=False)


def create_subcommands_parser():
    parser = argparse.ArgumentParser(
        prog="lb",
        description="xk media library",
        epilog="Report bugs here: https://github.com/chapmanjacobd/library/issues/new/choose",
        add_help=False,
    )
    subparsers = parser.add_subparsers()
    subp_fsadd = add_parser(subparsers, "fsadd", ["x", "extract"])
    subp_fsadd.set_defaults(func=fs_add)
    subp_fsupdate = add_parser(subparsers, "fsupdate", ["xu"])
    subp_fsupdate.set_defaults(func=fs_update)

    subp_watch = add_parser(subparsers, SC.watch, ["wt", "tubewatch", "tw", "entries"])
    subp_watch.set_defaults(func=watch)
    subp_listen = add_parser(subparsers, SC.listen, ["lt", "tubelisten", "tl"])
    subp_listen.set_defaults(func=listen)

    subp_read = add_parser(subparsers, SC.read, ["text", "books", "docs"])
    subp_read.set_defaults(func=read)
    subp_view = add_parser(subparsers, SC.view, ["image", "see", "look"])
    subp_view.set_defaults(func=view)

    subp_filesystem = add_parser(subparsers, SC.filesystem, ["fs"])
    subp_filesystem.set_defaults(func=filesystem)

    subp_bigdirs = add_parser(subparsers, "bigdirs", ["largefolders", "large_folders"])
    subp_bigdirs.set_defaults(func=scripts.bigdirs)
    subp_move_list = add_parser(subparsers, "mv-list", ["movelist", "move-list", "move_list"])
    subp_move_list.set_defaults(func=scripts.move_list)
    subp_relmv = add_parser(subparsers, "relmv", ["rel-mv", "mvrel", "mv-rel"])
    subp_relmv.set_defaults(func=scripts.relmv)
    subp_dedupe = add_parser(subparsers, "dedupe")
    subp_dedupe.set_defaults(func=scripts.dedupe)
    subp_scatter = add_parser(subparsers, "scatter")
    subp_scatter.set_defaults(func=scripts.scatter)
    subp_christen = add_parser(subparsers, "christen")
    subp_christen.set_defaults(func=scripts.christen)
    subp_merge_db = add_parser(subparsers, "merge-dbs")
    subp_merge_db.set_defaults(func=scripts.merge_dbs)
    subp_dedupe_local = add_parser(subparsers, "merge-online-local")
    subp_dedupe_local.set_defaults(func=scripts.merge_online_local)
    subp_optimize = add_parser(subparsers, "optimize")
    subp_optimize.set_defaults(func=scripts.optimize_db)

    subp_tubeadd = add_parser(subparsers, "tubeadd", ["dladd", "ta", "da", "xt"])
    subp_tubeadd.set_defaults(func=tube_add)
    subp_tubeupdate = add_parser(subparsers, "tubeupdate", ["dlupdate", "tu"])
    subp_tubeupdate.set_defaults(func=tube_update)

    subp_redditadd = add_parser(subparsers, "redditadd", ["ra", "xr"])
    subp_redditadd.set_defaults(func=reddit_add)
    subp_redditupdate = add_parser(subparsers, "redditupdate", ["ru", "xru"])
    subp_redditupdate.set_defaults(func=reddit_update)
    subp_pushshift = add_parser(subparsers, "pushshift", ["ps"])
    subp_pushshift.set_defaults(func=scripts.pushshift_extract)

    subp_hnadd = add_parser(subparsers, "hnadd")
    subp_hnadd.set_defaults(func=hacker_news_add)

    subp_download = add_parser(subparsers, "download", ["dl"])
    subp_download.set_defaults(func=dl_download)
    subp_block = add_parser(subparsers, "block")
    subp_block.set_defaults(func=dl_block)
    subp_redownload = add_parser(subparsers, "redownload", ["redl"])
    subp_redownload.set_defaults(func=scripts.redownload)

    subp_playlist = add_parser(subparsers, "playlists", ["pl", "folders"])
    subp_playlist.set_defaults(func=playlists)
    subp_dlstatus = add_parser(subparsers, "dlstatus", ["ds"])
    subp_dlstatus.set_defaults(func=dlstatus)
    subp_usage = add_parser(subparsers, "usage", ["du"])
    subp_usage.set_defaults(func=utils.mount_stats)

    subp_playback_now = add_parser(subparsers, "now")
    subp_playback_now.set_defaults(func=playback_now)
    subp_playback_next = add_parser(subparsers, "next")
    subp_playback_next.set_defaults(func=playback_next)
    subp_playback_stop = add_parser(subparsers, "stop")
    subp_playback_stop.set_defaults(func=playback_stop)
    subp_playback_pause = add_parser(subparsers, "pause")
    subp_playback_pause.set_defaults(func=playback_pause)

    subp_tabsadd = add_parser(subparsers, "tabsadd")
    subp_tabsadd.set_defaults(func=tabs_add)
    subp_tabs = add_parser(subparsers, "tabs", ["tb"])
    subp_tabs.set_defaults(func=tabs)
    subp_surf = add_parser(subparsers, "surf", ["browse", "load"])
    subp_surf.set_defaults(func=scripts.streaming_tab_loader)

    subp_nouns = add_parser(subparsers, "nouns")
    subp_nouns.set_defaults(func=scripts.nouns)

    subp_reddit_selftext = add_parser(subparsers, "reddit-selftext", ["rst"])
    subp_reddit_selftext.set_defaults(func=scripts.parse_reddit_selftext)
    subp_nfb_directors = add_parser(subparsers, "nfb-films")
    subp_nfb_directors.set_defaults(func=scripts.nfb_films)

    parser.add_argument("--version", "-V", action="store_true")
    return parser


def library(args=None) -> None:
    if args:
        sys.argv = ["lb"] + args

    parser = create_subcommands_parser()
    args, _unk = parser.parse_known_args(args)
    if args.version:
        return print(__version__)

    log.info("library v%s", __version__)
    log.info(sys.argv)
    original_argv = sys.argv
    if len(sys.argv) >= 2:
        del sys.argv[1]

    if hasattr(args, "func"):
        args.func()
    else:
        try:
            print("Subcommand", original_argv[1], "not found")
        except Exception:
            print("Invalid args. I see:", original_argv)

        print_help(parser)


if __name__ == "__main__":
    library()



import re

reddit_links_ignore = re.compile(
    "|".join(
        r""".*youtube.com/user/
.*youtube.com/c/
.*youtube.com/channel/
.*youtube.com/results""".splitlines()
    )
)

yt_recoverable_errors = re.compile(
    "|".join(
        r""".*due to geo restriction
.*geo-restricted
.*geolocation
.*your country
.*No such file or directory
.*The downloaded file is empty
.*Invalid data found
.*fragment 1 not found
.*HTTP Error 429
.*HTTP Error 400
.*HTTP Error 503: Service Unavailable
.*Origin Error
.*API is not granting access
.*Did not get any data blocks
.*Too Many Requests
.*Postprocessing
.*Premieres in
.*user.*not allowed
.*read operation timed out
.*Could not connect
.*restricted
.*Internal Server Error
.*Internal error encountered
.*not currently available
.*currently unavailable
.*selectionunavailable
.*Downloaded .* bytes, expected .* bytes
.*Playlists that require authentication may not extract correctly without a successful webpage download
.*Main webpage is locked behind the login page
.*You need to log in to access this content
.*This video is only available for registered users
.*members-only content
.*not available in your location
.*linked to an account
.*This video is only available to Music Premium members
.*episode is not currently available
.*This channel has no uploads
.*Could not send HEAD request
.*Unable to download JSON metadata
.*Failed to parse JSON Expecting
.*expected string or bytes-like object
.*Connection refused
.*Failed to download MPD manifest:
.*not currently available
.*copyright claim""".splitlines()
    )
)


yt_meaningless_errors = re.compile(
    "|".join(
        r"""^\[info\]
^\[redirect\]
^\[Merger\]
^\[dashsegments\]
^Found
.*hidden
.*timed out
.*Timeout
.*Timed
.*Connection reset
.*ConnectionReset
.*Unable to extract
.*Unauthorized
.*Forbidden
.*Traceback
.*Invalid argument
.*KeyboardInterrupt
.*Fatal Python error
.*list index out of range
.*Extract.* cookies
.*File .*, line .*, in
.*Requested format is not available.
.*This channel does not have
.*fragment_filename_sanitized
.*no suitable InfoExtractor for URL
.*File name too long
.*No such file or directory
.*Name or service not known
.*: Downloading webpage
.*: Extracting information
.*: Requesting header
.*mismatched tag
.*Fail.* parse
.*Downloading .* metadata
.*Downloading .* information
.*Downloading .* manifest
.*Determining source extension
.*Downloading jwt token
.*Skipping embedding .* subtitle because the file is missing
.*Finished downloading playlist
.*The last 30x error message was:
.*NoneType
.*unable to download video data
.*HTTP Error 405
.*Creating a generic title instead
.*The channel is not currently live
.*clips are not currently supported.
.*Confirm you are on the latest version using
.*referenced before assignment
.*field is missing or empty
.*list index out of range
.*Interrupted by user
.*unable to open for writing:
.*You might want to use a VPN or a proxy server
.*maximum recursion depth exceeded
.*object does not support item assignment
.*encodings are not supported
.*object has no attribute
.*merged
.*Compressed file ended before the end-of-stream marker was reached
.*Falling back on generic
.*Some formats are possibly damaged
.*WARNING: unable to obtain file audio codec with ffprobe
.*matching opening tag for closing p tag not found
.*the JSON object must be str, bytes or bytearray, not dict
.*list indices must be integers
.*The read operation timed out
.*Unable to recognize playlist.
.*Premieres in""".splitlines()
    )
)

yt_unrecoverable_errors = re.compile(
    "|".join(
        r""".*repetitive or misleading metadata
.*It is not available
.*has already been recorded in the archive
.*ideo.*is private
.*No video player ID
.*has been removed
.*from a suspended account
.*has not been found
.*This video has been disabled
.*No suitable extractor
.*The URL must be suitable for the extractor
.*Private video
.*not a video
.*content expired
.*Invalid URL
.*Incomplete.*ID
.*The given url does not contain a video
.*because this account owner limits who can view
.*No status found with that ID
.*program functionality for this site has been marked as broken, and will probably not work
.*This playlist is private
.*Unable to recognize tab page
.*'bytes' object has no attribute 'encode'
.*only video extraction is supported
.*'NoneType' object has no attribute 'get'
.*list index out of range
.*Unable to extract cnn url
.*PornHd.*Unable to extract error message
.*status error-notFound
.*ERROR.*items
.*You don't have permission to access this video.
.*Video is unavailable pending review
.*Video has been flagged for verification
.*This video has been disabled
.*The uploader has not made this video available.
.*channel/playlist does not exist
.*This video is DRM protected
.*This video is protected by a password
.*This video requires payment to watch.
.*Unable to download webpage
.*dashboard-only post
.*The policy key provided does not permit this account or video
.*live stream recording
.*The channel is not currently live
.*This live event will begin in a few moments
.*nudity or sexual content
.*policy on harassment and bullying
.*Can't find object userchannel
.*is offline
.*: Video unavailable
.*does not exist
.*has been removed
.*no video on the webpage
.*Track not found
.*: Not found.
.*Video no longer exists
.*Premieres in.*hours
.*This clip is no longer available
.*No media found
.*No sources found for video
.*caused by KeyError\('title'
.*\[youtube:truncated_url\]
.*: This video is not available.
.*Video unavailable. This video is not available
.*This article does not contain a video
.*Resource temporarily unavailable
.*This video is unavailable
.*No video could be found in this
.*Unsupported URL
.*The requested site is known to use DRM protection.
.*URL could be a direct video link
.*No media found
.*not a valid URL
.*not a video
.*The page doesn't contain any tracks
.*removed by the uploader
.*Access to this video has been restricted by its creator
.*blocked due to author's rights infingement
.*blocked it on copyright grounds
.*uploader has closed their.*account
.*account has been terminated because we received multiple
.*User has been suspended
.*policy on violent or graphic content
.*The channel does not have a .* tab
.*policy on spam, deceptive practices, and scams
.*This video does not exist, or has been deleted.
.*The playlist does not exist.
.*Community Guidelines
.*Terms of Service
.*This channel does not exist
.*account that no longer exists
.*account associated with this video
.*This video doesn't exist.
.*Track not found
.*Not found.
.*Can't find object media for
.*No video formats found
.*certificate is not valid
.*CERTIFICATE_VERIFY_FAILED
.*HTTP Error 403: Forbidden
.*code -404
.*HTTP Error 404
.*HTTPError 404
.*HTTP Error 410
.*HTTPError 410""".splitlines()
    )
)


prefix_unrecoverable_errors = re.compile(
    "|".join(
        r""".*unable to write data:
.*No space left on device
.*Transport endpoint is not connected
.*unable to create directory
.*Permission denied""".splitlines()
    )
)



from scripts.bigdirs import bigdirs, process_bigdirs
from scripts.christen import christen
from scripts.dedupe import dedupe
from scripts.merge_dbs import merge_dbs
from scripts.merge_online_local import merge_online_local
from scripts.mining.nfb_ca import nfb_films
from scripts.mining.nouns import nouns
from scripts.mining.pushshift import pushshift_extract
from scripts.mining.reddit_self import parse_reddit_selftext
from scripts.move_list import move_list
from scripts.optimize_db import optimize_db
from scripts.redownload import redownload
from scripts.relmv import relmv
from scripts.scatter import scatter, scatter_usage
from scripts.streaming_tab_loader import streaming_tab_loader



import argparse, html
from urllib.parse import urlparse

from xklb import db, utils
from xklb.utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("database")
    parser.add_argument("--verbose", "-v", action="count", default=0)
    args = parser.parse_args()

    args.db = db.connect(args)
    log.info(utils.dict_filter_bool(args.__dict__))
    return args


def get_page_links(path, text):
    from bs4 import BeautifulSoup

    soup = BeautifulSoup(html.unescape(text), "lxml")
    internal_links = set()
    external_links = set()

    for a in soup.findAll("a", attrs={"href": True}):
        href = a["href"].strip()
        if all([len(href) > 1, href[0] != "#", "javascript:" not in href, "mailto:" not in href, "tel:" not in href]):
            if "http" in href or "https" in href:
                if urlparse(path).netloc.lower() in urlparse(href).netloc.lower():
                    internal_links.add(a["href"])
                else:
                    external_links.add(a["href"])
            else:
                utils.log.debug(a["href"])

    return internal_links, external_links


def parse_reddit_selftext() -> None:
    from markdown import markdown

    args = parse_args()
    m_columns = args.db["media"].columns_dict

    reddit_posts = list(
        args.db.query(
            f"""
            select path, selftext
            from reddit_posts
            where 1=1
                {'AND path not in (select distinct webpage from media where webpage is not null)' if 'webpage' in m_columns else ''}
            """
        )
    )

    for d in reddit_posts:
        html_data = markdown(d["selftext"])
        internal_links, external_links = get_page_links(d["path"], html_data)
        if internal_links:
            for i_link in internal_links:
                log.info(i_link)

        for e_link in external_links:
            args.db["media"].upsert({**d, "path": e_link, "webpage": d["path"]}, pk="path", alter=True)


if __name__ == "__main__":
    parse_reddit_selftext()
